﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Security.Cryptography;
using Lab2DiffieHellman;
using System.IO;

namespace Lab2DHServer
{
    public class ServerListener
    {

        public static string data = null;
        static int port = 11001;
        
        /// <summary>
        /// This method is used to start listening server on preset port, establish the connection, exchange key and exchange messages.
        /// </summary>
        public static void StartListening()
        {
            byte[] bytes = new Byte[1024];

            IPHostEntry ipHostName = Dns.GetHostEntry(Dns.GetHostName());
            IPAddress ipAdd = ipHostName.AddressList.Last(t => t.AddressFamily == AddressFamily.InterNetwork);
            
            TcpListener listener = new TcpListener(IPAddress.Any, port);
            listener.Start();

            try
            {
                Console.WriteLine("Waiting for a connection at the port {0}...", port);
                TcpClient client = listener.AcceptTcpClient();

                Console.WriteLine("Connection from client accepted.");
                NetworkStream stream = client.GetStream();

                byte[] key = ExchangeKey(stream);
                
                while (true)
                {
                    string reply = GetMessage(stream, key);

                    SendReply(stream, reply, key);
                }


            }
            catch (Exception)
            {
                Console.WriteLine("Connection lost.");
            }
            finally
            {
                listener.Stop();
            }

            Console.WriteLine("\nDo you want to keep listening? (y/n)");
            string keepGoing = Console.ReadLine();
            if (keepGoing == "y") StartListening();
        }

        /// <summary>
        /// This method is used for encrypting and sending replies to the client.
        /// </summary>
        /// <param name="stream">NetworkStream to the client.</param>
        /// <param name="reply">String with the message to be sent.</param>
        /// <param name="key">Exchanged private key.</param>
        private static void SendReply(NetworkStream stream, string reply, byte[] key)
        {
            byte[] outgoing;
            using (Aes encryption = Aes.Create())
            {
                encryption.Key = key;
                encryption.GenerateIV();

                byte[] mess = AesTools.EncryptStringToBytes(reply, encryption);
                outgoing = new byte[encryption.IV.Length + mess.Length];
                encryption.IV.CopyTo(outgoing, 0);
                mess.CopyTo(outgoing, encryption.IV.Length);
            }

            try
            {
                stream.Write(outgoing, 0, outgoing.Length);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        /// <summary>
        /// This method is used for receiving messages from client, decoding them and processing them.
        /// </summary>
        /// <param name="stream">NetworkStream to the client.</param>
        /// <param name="key">Exchanged secret key.</param>
        /// <returns>Processed decoded message from the client.</returns>
        private static string GetMessage(NetworkStream stream, byte[] key)
        {
            byte[] incoming = new byte[1024];
            int bytesRead = stream.Read(incoming, 0, incoming.Length);
            byte[] trimmed = new byte[bytesRead - (16)];
            byte[] iv = new byte[16];

            for (int i = 0; i < iv.Length; i++)
            {
                iv[i] = incoming[i];
            }

            for (int i = iv.Length; i < bytesRead; i++)
            {
                trimmed[i - iv.Length] = incoming[i];
            }

            string reply;
            using (Aes encryption = Aes.Create())
            {
                encryption.Key = key;
                encryption.IV = iv;

                string received = AesTools.DecryptStringFromBytes(trimmed, encryption);
                reply = string.Format("I've received \"{0}\"", received);
                Console.WriteLine("The received message is: {0}", received);
            }
            return reply;
        }

        /// <summary>
        /// This method is used to exchange secret key with the client after establishing the connection.
        /// </summary>
        /// <param name="stream">NetworkStream to the client.</param>
        /// <returns>Exchanged secret key.</returns>
        private static byte[] ExchangeKey(NetworkStream stream)
        {
            byte[] incoming = new byte[16384];
            int bytesRead = stream.Read(incoming, 0, incoming.Length);

            DifHel DHServer = new DifHel(Encoding.Unicode.GetString(incoming));

            try
            {
                stream.Write(Encoding.Unicode.GetBytes(DHServer.ToString()), 0, Encoding.Unicode.GetByteCount(DHServer.ToString()));
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }

            Console.WriteLine("The key was succesfully calculated.");
            return DHServer.Key;
        }
    }
}
